import extractor
import unittest


class ExtractUfoTest(unittest.TestCase):

    # TODO: replace this with real test cases
    def test_extractUFO_dummy_test(self):
        self.assertTrue(True)


if __name__ == "__main__":
    unittest.main()
